from sklearn.linear_model import LinearRegression
from sklearn.model_selection import train_test_split
from cassandra.data.trasformations.trasformations import create_model
from cassandra.model.modelEvaluation.plot import show_nrmse, show_mape, show_rsquared, show_rssd
import numpy as np


def linear(df, X_columns, target, name_model, medias=[], organic=[], metric=None,
           return_metric=False, size=0.2, positive=False, random_state=42, force_coeffs=False, coeffs=[], intercept=0):
    if metric is None:
        metric = ['rsq_train', 'rsq_test', 'nrmse_train', 'nrmse_test', 'mape_train', 'mape_test', 'rssd']

    X = df[X_columns]
    y = df[target]

    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=size, random_state=random_state)

    if medias or organic:
        model = create_model(medias, organic, LinearRegression(positive=positive))
    else:
        model = LinearRegression(positive=positive)

    model.fit(X_train, y_train)

    if force_coeffs:
        model.intercept_ = intercept
        if coeffs:
            model.coef_ = np.array(coeffs)

    # Ask the model to predict on X_test without having Y_test
    # This will give you exact predicted values

    # We can use our NRMSE and MAPE functions as well

    # Create new DF not to edit the original one
    result = df

    # Create a new column with predicted values
    if medias or organic:
        result['prediction'] = model.predict(result)
        y_train_pred = model.predict(X_train)
        y_test_pred = model.predict(X_test)
    else:
        result['prediction'] = model.predict(X)
        y_train_pred = model.predict(X_train)
        y_test_pred = model.predict(X_test)

    #metrics_values = choose_metric(name_model, metric, return_metric, y_train, y_test, y_train_pred, y_test_pred)
    metrics_values = {}

    # Score returns the accuracy of the above prediction or R^2
    if 'rsq_train' in metric:
        try:
            rsq_train = show_rsquared(np.array(y_train), np.array(y_train_pred))
        except:
            rsq_train = -100
        if return_metric:
            metrics_values[name_model + '_rsq_train'] = rsq_train
        print(name_model, 'RSQ train: ', rsq_train)

    if 'rsq_test' in metric:
        try:
            rsq_test = show_rsquared(np.array(y_test), np.array(y_test_pred))
        except:
            rsq_test = -100
        if return_metric:
            metrics_values[name_model + '_rsq_test'] = rsq_test
        print(name_model, 'RSQ test: ', rsq_test)

    # Get the NRMSE values
    if 'nrmse_train' in metric:
        try:
            nrmse_train_val = show_nrmse(np.array(y_train), np.array(y_train_pred))
        except:
            nrmse_train_val = 100
        if return_metric:
            metrics_values[name_model + '_nrmse_train'] = nrmse_train_val
        print(name_model, 'NRMSE train: ', nrmse_train_val)

    if 'nrmse_test' in metric:
        try:
            nrmse_test_val = show_nrmse(np.array(y_test), np.array(y_test_pred))
        except:
            nrmse_test_val = 100
        if return_metric:
            metrics_values[name_model + '_nrmse_test'] = nrmse_test_val
        print(name_model, 'NRMSE test: ', nrmse_test_val)

    # Get the MAPE values
    if 'mape_train' in metric:
        try:
            mape_train_val = show_mape(np.array(y_train), np.array(y_train_pred))
        except:
            mape_train_val = 100
        if return_metric:
            metrics_values[name_model + '_mape_train'] = mape_train_val
        print(name_model, 'MAPE train: ', mape_train_val)

    if 'mape_test' in metric:
        try:
            mape_test_val = show_mape(np.array(y_test), np.array(y_test_pred))
        except:
            mape_test_val = 100
        if return_metric:
            metrics_values[name_model + '_mape_test'] = mape_test_val
        print(name_model, 'MAPE test: ', mape_test_val)

    if 'rssd' in metric:
        try:
            rssd_val = show_rssd(X, model.coef_)
        except:
            rssd_val = 100
        if return_metric:
            metrics_values[name_model + '_rssd'] = rssd_val
        print(name_model, 'RSSD: ', rssd_val)

    if metrics_values:
        return result, model, metrics_values
    else:
        return result, model


# def choose_metric(name_model, metric, return_metric, y_train, y_test, y_train_pred, y_test_pred):
#     metrics_values = {}
#
#     # Score returns the accuracy of the above prediction or R^2
#     if 'rsq_train' in metric:
#         try:
#             rsq_train = show_rsquared(np.array(y_train), np.array(y_train_pred))
#         except:
#             rsq_train = -100
#         if return_metric:
#             metrics_values[name_model + '_rsq_train'] = rsq_train
#         print(name_model, 'RSQ train: ', rsq_train)
#
#     if 'rsq_test' in metric:
#         try:
#             rsq_test = show_rsquared(np.array(y_test), np.array(y_test_pred))
#         except:
#             rsq_test = -100
#         if return_metric:
#             metrics_values[name_model + '_rsq_test'] = rsq_test
#         print(name_model, 'RSQ test: ', rsq_test)
#
#     # Get the NRMSE values
#     if 'nrmse_train' in metric:
#         try:
#             nrmse_train_val = show_nrmse(np.array(y_train), np.array(y_train_pred))
#         except:
#             nrmse_train_val = 100
#         if return_metric:
#             metrics_values[name_model + '_nrmse_train'] = nrmse_train_val
#         print(name_model, 'NRMSE train: ', nrmse_train_val)
#
#     if 'nrmse_test' in metric:
#         try:
#             nrmse_test_val = show_nrmse(np.array(y_test), np.array(y_test_pred))
#         except:
#             nrmse_test_val = 100
#         if return_metric:
#             metrics_values[name_model + '_nrmse_test'] = nrmse_test_val
#         print(name_model, 'NRMSE test: ', nrmse_test_val)
#
#     # Get the MAPE values
#     if 'mape_train' in metric:
#         try:
#             mape_train_val = show_mape(np.array(y_train), np.array(y_train_pred))
#         except:
#             mape_train_val = 100
#         if return_metric:
#             metrics_values[name_model + '_mape_train'] = mape_train_val
#         print(name_model, 'MAPE train: ', mape_train_val)
#
#     if 'mape_test' in metric:
#         try:
#             mape_test_val = show_mape(np.array(y_test), np.array(y_test_pred))
#         except:
#             mape_test_val = 100
#         if return_metric:
#             metrics_values[name_model + '_mape_test'] = mape_test_val
#         print(name_model, 'MAPE test: ', mape_test_val)
#
#     return metrics_values



import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.validation import check_is_fitted, check_array
from scipy.signal import convolve2d
from sklearn.compose import ColumnTransformer
from sklearn.pipeline import Pipeline
from sklearn.impute import SimpleImputer
import statsmodels.tsa.api as tsa
from scipy.stats import dweibull


# Define the Adstock Class as a custom Transformer
class ExponentialCarryover(BaseEstimator, TransformerMixin):
    def __init__(self, strength=0.5, length=1):
        self.strength = strength
        self.length = length

    def fit(self, X, y=None):
        X = check_array(X)
        self._check_n_features(X, reset=True)
        self.sliding_window_ = (
                self.strength ** np.arange(self.length + 1)
        ).reshape(-1, 1)

        return self

    def transform(self, X: np.ndarray):
        check_is_fitted(self)
        X = check_array(X)
        self._check_n_features(X, reset=False)
        convolution = convolve2d(X, self.sliding_window_)

        if self.length > 0:
            convolution = convolution[: -self.length]

        return convolution


# Define the Saturation Class as a custom Transformer
class ExponentialSaturation:
    def __init__(self, a=1.):
        self.a = a

    def transform(self, X):
        return 1 - np.exp(-self.a * X)


class ExponentialSaturation(BaseEstimator, TransformerMixin):
    def __init__(self, a=1.):
        self.a = a

    def fit(self, X, y=None):
        X = check_array(X)
        self._check_n_features(X, reset=True)  # from BaseEstimator

        return self

    def transform(self, X):
        check_is_fitted(self)
        X = check_array(X)
        self._check_n_features(X, reset=False)  # from BaseEstimator

        return 1 - np.exp(-self.a * X)


def trasformation(media, organic):
    trasf = []
    for x in media:
        trasf.append(
            (x, Pipeline([
                ('carryover', ExponentialCarryover()),
                ('saturation', ExponentialSaturation())
            ]), [x]),
        )
    for x in organic:
        trasf.append(
            (x, Pipeline([
                ('inputer', SimpleImputer())
            ]), [x]),
        )
    return trasf


def create_model(medias, organic, model):
    trasf = ColumnTransformer(
        trasformation(medias, organic)
    )

    steps = [
        ('trasformation', trasf),
        ('regression', model)
    ]
    pipeline = Pipeline(steps)

    return pipeline


# Function to return Adstocked variables
def adstock_geometric_old(x, theta):
    return tsa.filters.recursive_filter(x, theta)


# Function to return Adstock weibull pdf variables
def adstock_weibull_pdf(x, scale, shape):
    return (shape / scale) * (x / scale) ** (shape - 1) * np.exp(-(x / scale) ** shape)


# Function to return Adstock weibull cdf variables
def adstock_weibull_cdf(x, scale, shape):
    return 1 - np.exp(-(x / scale) ** shape)


# Function to return Saturated variables
def saturation(x, beta):
    return x ** beta


def saturation_robyn(x, coeff, alpha, gamma):
    return coeff * (x ** alpha / (x ** alpha + gamma ** alpha))


def saturation_hill(x, alpha, gamma):
    if isinstance(x, list):
        gammaTrans = round(np.percentile(np.linspace(start=min(x), stop=max(x), num=100), 100 * gamma), 4)
        x_scurve = x ** alpha / (x ** alpha + gammaTrans ** alpha)
    else:
        x_scurve = x ** alpha / (x ** alpha + gamma ** alpha)
    return x_scurve


def adstock_geometric(x, theta):
    x_decayed = np.concatenate(([x[1]], np.repeat(0, len(x) - 1)))
    for xi in range(1, len(x_decayed)):
        x_decayed[xi] = x[xi] + theta * x_decayed[xi - 1]

    thetaVecCum = []
    thetaVecCum.insert(0, theta)
    for t in range(1, len(x)):
        thetaVecCum.insert(t, thetaVecCum[t - 1] * theta)

    all = {'x': x, 'x_decayed': x_decayed, 'thetaVecCum': thetaVecCum}

    return pd.Series(x_decayed)


def adstock_weibull(x, scale, shape, weibull_type="pdf"):
    windlen = len(x)
    x_bin = [*range(1, windlen + 1)]
    scaleTrans = round(np.percentile(x_bin, 100 * scale), 0)

    if weibull_type == 'pdf':

        def normalize(x):
            decimal_x = x.iloc[:, 0].apply(float)

            if (max(decimal_x) - min(decimal_x)) == 0:
                return np.concatenate(([1], np.repeat(0, windlen - 1)))
            else:
                return (decimal_x - min(decimal_x)) / (max(decimal_x) - min(decimal_x))

        thetaVecCum = normalize(adstock_weibull_pdf(pd.DataFrame(x_bin), scaleTrans, shape))
        thetaVecCumArray = np.array(thetaVecCum.values.tolist())

    # elif weibull_type == 'cdf':
    # thetaVec = np.concatenate((1, 1 - pweibull(head(x_bin, -1), shape = shape, scale = scaleTrans)) # plot(thetaVec)
    # thetaVecCum <- cumprod(thetaVec) # plot(thetaVecCum)

    def x_decayed(x_val, x_pos):
        x_vec = np.concatenate((np.repeat(0, x_pos - 1), np.repeat(x_val, windlen - x_pos + 1)))
        thetaVecCumLag = np.roll(thetaVecCumArray, x_pos - 1)
        # shift(thetaVecCum, x_pos - 1, fill = 0)
        x_prod = x_vec * thetaVecCumLag
        thetaVecCumLag[x_pos - 1] = 0
        return x_prod

    x_decay = np.sum(list(map(x_decayed, x, x_bin)), axis=1)

    all = {'x': np.array(x), 'x_decayed': np.array(x_decay[::-1]), 'thetaVecCum': thetaVecCum.values.tolist()}

    return pd.Series(x_decay[::-1])



import pandas as pd
import numpy as np
from scipy.stats import dweibull
import matplotlib.pyplot as plt
import plotly.express as px


def create_df_adstock_weibull(df, adstock_type, weibull_type='pdf', max_y=1, max_x=100, graph=False):
    df_dict = df.to_dict()
    new_dict = {}
    for key, value in df_dict.items():
        new_dict[key] = value['Initial Model']

    if adstock_type == 'weibull':
        result_df = pd.DataFrame()
        adstock_columns = [col.replace('_shapes', '') for col in new_dict.keys() if 'shapes' in col]

        for m in adstock_columns:
            shape_name = m + '_shapes'
            scale_name = m + '_scales'
            shape = new_dict[shape_name]
            scale = new_dict[scale_name]

            x_array = np.array([1])
            x = np.repeat(x_array, 100, axis=0)  # see to del
            range_x = [x / 100 for x in range(1, 101)]

            if weibull_type == 'pdf':
                y = dweibull.pdf(range_x, shape, scale=scale)
            else:
                y = dweibull.cdf(range_x, shape, scale=scale)

            data = plt.plot(range(1, 101), y)[0].get_data()

            df_media = pd.DataFrame(data).T

            df_media.rename(columns={df_media.columns[0]: 'x', df_media.columns[1]: 'y'},
                            inplace=True)

            for index, row in df_media.iterrows():
                df_media.at[index, 'y'] = ((df_media.at[index, 'y'] - df_media['y'].min()) / (
                        df_media['y'].max() - df_media['y'].min())) * (max_y - 0) + 0
                df_media.at[index, 'x'] = ((df_media.at[index, 'x'] - df_media['x'].min()) / (
                        df_media['x'].max() - df_media['x'].min())) * (max_x - 0) + 0
            if graph == True:
                title_graph = 'Adstock weibull ' + m
                fig2 = px.line(x=df_media['x'], y=df_media['y'], title=title_graph)
                fig2.show()

            df_media['canale'] = m
            result_df = pd.concat([result_df, df_media])

        shapes = [col for col in new_dict.keys() if 'shapes' in col]
        adstocked_shapes = {key: new_dict[key] for key in shapes}
        scales = [col for col in new_dict.keys() if 'scales' in col]
        adstocked_scales = {key: new_dict[key] for key in scales}

        adstock_shapes_df = pd.DataFrame(list(adstocked_shapes.items()))
        adstock_shapes_df.rename(
            columns={adstock_shapes_df.columns[0]: 'canale', adstock_shapes_df.columns[1]: 'adstock_shapes'},
            inplace=True)
        adstock_shapes_df['canale'] = adstock_shapes_df.canale.str.replace('_shapes', '')

        adstock_scales_df = pd.DataFrame(list(adstocked_scales.items()))
        adstock_scales_df.rename(
            columns={adstock_scales_df.columns[0]: 'canale', adstock_scales_df.columns[1]: 'adstock_scales'},
            inplace=True)
        adstock_scales_df['canale'] = adstock_scales_df.canale.str.replace('_scales', '')

        adstock_df = pd.merge(left=adstock_shapes_df, right=adstock_scales_df, left_on='canale',
                              right_on='canale')

    if adstock_type == 'weibull':
        return adstock_df, result_df


def create_df_saturation(df_hyperparameters, df_spend_response, df_aggregated, graph=True):
    df_dict = df_hyperparameters.to_dict()
    new_dict = {}
    for key, value in df_dict.items():
        new_dict[key] = value['Initial Model']

    result_df = pd.DataFrame()
    saturation_columns = list(df_spend_response['canale'].unique())

    for m in saturation_columns:
        alpha_name = m + '_alphas'
        gamma_name = m + '_gammas'
        alpha = new_dict[alpha_name]
        gamma = new_dict[gamma_name]
        coef = df_aggregated['coef'].copy().loc[df_aggregated['rn'] == m][0]

        range_x = list(df_spend_response['spend'].copy().loc[df_spend_response['canale'] == m])
        max_x = max(range_x)
        min_x = min(range_x)

        range_y = list(df_spend_response['response'].copy().loc[df_spend_response['canale'] == m])
        max_y = max(range_y)

        gamma_trans = round(np.quantile(np.linspace(start=min_x, stop=max_x, num=100), gamma), 4)

        y_res = []
        for x in range_x:
            y_res.append((x ** alpha / (x ** alpha + gamma_trans ** alpha)) * coef)

        max_y_res = max(y_res)
        min_y_res = min(y_res)

        data = plt.plot(range_x, y_res)[0].get_data()
        df_new = pd.DataFrame(data).T
        df_new.rename(columns={df_new.columns[0]: 'x', df_new.columns[1]: 'y'},
                      inplace=True)

        for index, row in df_new.iterrows():
            df_new.at[index, 'y'] = ((df_new.at[index, 'y'] - df_new['y'].min()) / (
                    df_new['y'].max() - df_new['y'].min())) * (max_y_res - min_y_res) + min_y_res
            df_new.at[index, 'x'] = ((df_new.at[index, 'x'] - df_new['x'].min()) / (
                    df_new['x'].max() - df_new['x'].min())) * (max_x - min_x) + min_x

        if graph == True:
            title_graph = 'Saturation ' + m
            fig2 = px.line(x=df_new['x'], y=df_new['y'], title=title_graph)

            fig2.update_layout(
                xaxis_title=m + ' spend',
                yaxis_title=m + ' response')

            # TODO
            # fig2.add_trace(
            # go.Scatter(
            # mode='markers',
            # x=[df_new['x'].mean()],
            # marker=dict(
            # size=12),
            # name='mean spend'
            # )
            # )
            fig2.show()

        df_new['canale'] = m
        result_df = pd.concat([result_df, df_new])

    return result_df

def new_create_df_adstock_weibull(df, adstock_type, weibull_type='pdf', max_y=1, max_x=100, graph=False):
    df_dict = df.to_dict()
    new_dict = {}
    for key, value in df_dict.items():
        new_dict[key] = value['Initial Model']

    if adstock_type == 'weibull':
        result_df = pd.DataFrame()
        adstock_columns = [col.replace('_shapes', '') for col in new_dict.keys() if 'shapes' in col]

        for m in adstock_columns:
            shape_name = m + '_shapes'
            scale_name = m + '_scales'
            shape = new_dict[shape_name]
            scale = new_dict[scale_name]

            x_array = np.array([1])
            x = np.repeat(x_array, 100, axis=0)  # see to del
            range_x = [x / 100 for x in range(1, 101)]

            if weibull_type == 'pdf':
                y = dweibull.pdf(range_x, shape, scale=scale)
            else:
                y = dweibull.cdf(range_x, shape, scale=scale)

            data = plt.plot(range(1, 101), y)[0].get_data()

            df_media = pd.DataFrame(data).T

            df_media.rename(columns={df_media.columns[0]: 'x', df_media.columns[1]: 'y'},
                            inplace=True)

            for index, row in df_media.iterrows():
                df_media.at[index, 'y'] = ((df_media.at[index, 'y'] - df_media['y'].min()) / (
                        df_media['y'].max() - df_media['y'].min())) * (max_y - 0) + 0
                df_media.at[index, 'x'] = ((df_media.at[index, 'x'] - df_media['x'].min()) / (
                        df_media['x'].max() - df_media['x'].min())) * (max_x - 0) + 0
            if graph == True:
                title_graph = 'Adstock weibull ' + m
                fig2 = px.line(x=df_media['x'], y=df_media['y'], title=title_graph)
                fig2.show()

            df_media['canale'] = m
            result_df = pd.concat([result_df, df_media])

        shapes = [col for col in new_dict.keys() if 'shapes' in col]
        adstocked_shapes = {key: new_dict[key] for key in shapes}
        scales = [col for col in new_dict.keys() if 'scales' in col]
        adstocked_scales = {key: new_dict[key] for key in scales}

        adstock_shapes_df = pd.DataFrame(list(adstocked_shapes.items()))
        adstock_shapes_df.rename(
            columns={adstock_shapes_df.columns[0]: 'canale', adstock_shapes_df.columns[1]: 'adstock_shapes'},
            inplace=True)
        adstock_shapes_df['canale'] = adstock_shapes_df.canale.str.replace('_shapes', '')

        adstock_scales_df = pd.DataFrame(list(adstocked_scales.items()))
        adstock_scales_df.rename(
            columns={adstock_scales_df.columns[0]: 'canale', adstock_scales_df.columns[1]: 'adstock_scales'},
            inplace=True)
        adstock_scales_df['canale'] = adstock_scales_df.canale.str.replace('_scales', '')

        adstock_df = pd.merge(left=adstock_shapes_df, right=adstock_scales_df, left_on='canale',
                              right_on='canale')

    if adstock_type == 'weibull':
        return adstock_df, result_df



from cassandra.data.trasformations.trasformations import adstock_weibull, saturation, saturation_hill, adstock_geometric
import nevergrad as ng
import numpy as np


def create_df_transformations_nevergrad(trasf, df, df_transformations, medias, use_adstock=True, use_saturation=True,
                                        adstock_type='weibull', saturation_type='hill'):
    if use_saturation:
        if saturation_type == 'hill':
            if use_adstock:
                if adstock_type == 'weibull':
                    for m in medias:
                        df_transformations[m] = saturation_hill(
                            adstock_weibull(df[m], trasf[m + '_scale'], trasf[m + '_shape']).replace([np.inf, -np.inf],
                                                                                                     0),
                            trasf[m + '_alpha'], trasf[m + '_gamma'])

                elif adstock_type == 'geometric':
                    for m in medias:
                        df_transformations[m] = saturation_hill(adstock_geometric(df[m], trasf[m + '_theta']),
                                                                trasf[m + '_alpha'], trasf[m + '_gamma'])
            else:
                for m in medias:
                    df_transformations[m] = saturation_hill(df[m], trasf[m + '_alpha'], trasf[m + '_gamma'])

        else:
            if use_adstock:
                if adstock_type == 'weibull':
                    for m in medias:
                        df_transformations[m] = saturation(
                            adstock_weibull(df[m], trasf[m + '_scale'], trasf[m + '_shape']).replace([np.inf, -np.inf],
                                                                                                     0),
                            trasf[m + '_beta'])

                elif adstock_type == 'geometric':
                    for m in medias:
                        df_transformations[m] = saturation(adstock_geometric(df[m], trasf[m + '_theta']), trasf[m + '_beta'])
            else:
                for m in medias:
                    df_transformations[m] = saturation(df[m], trasf[m + '_beta'])

    else:
        if use_adstock:
            if adstock_type == 'weibull':
                for m in medias:
                    df_transformations[m] = adstock_weibull(df[m], trasf[m + '_scale'], trasf[m + '_shape']).replace(
                        [np.inf, -np.inf], 0)

            elif adstock_type == 'geometric':
                for m in medias:
                    df_transformations[m] = adstock_geometric(df[m], trasf[m + '_theta'])

    return df_transformations


def instrum_variables_nevergrad(medias, organic, force_coeffs=False, use_intercept=True, use_adstock=True,
                                use_saturation=True, saturation_type='hill', adstock_type='weibull'):
    result_dict = {}

    if force_coeffs:
        for x in medias:
            result_dict[x] = {'lower': 0, 'upper': None}
        for x in organic:
            if '_org' in x:
                result_dict[x] = {'lower': 0, 'upper': None}
            else:
                result_dict[x] = {'lower': None, 'upper': None}
        if use_intercept:
            result_dict['intercept'] = {'lower': None, 'upper': None}

    if use_adstock:
        if adstock_type == 'weibull':
            for x in medias:
                result_dict[x + '_shape'] = {'lower': 0.0001, 'upper': 10}
                result_dict[x + '_scale'] = {'lower': 0, 'upper': 0.1}

                if use_saturation:
                    if saturation_type == 'hill':
                        result_dict[x + '_alpha'] = {'lower': 0.5, 'upper': 3}
                        result_dict[x + '_gamma'] = {'lower': 0.3, 'upper': 1}
                    else:
                        result_dict[x + '_beta'] = {'lower': 0.5, 'upper': 3}

        elif adstock_type == 'geometric':
            for x in medias:
                result_dict[x + '_theta'] = {'lower': 0, 'upper': 0.3}

                if use_saturation:
                    if saturation_type == 'hill':
                        result_dict[x + '_alpha'] = {'lower': 0.5, 'upper': 3}
                        result_dict[x + '_gamma'] = {'lower': 0.3, 'upper': 1}
                    else:
                        result_dict[x + '_beta'] = {'lower': 0.5, 'upper': 3}

    return result_dict


def instrum_transformations_nevergrad(hyperparameters_dict):
    result = {}
    for key, value in hyperparameters_dict.items():
        result[key] = ng.p.Scalar(lower=hyperparameters_dict[key]['lower'], upper=hyperparameters_dict[key]['upper'])

    return result


# All algorithms have strengths and weaknesses. Questionable rules of thumb could be: @NGOpt is “meta”-optimizer which
# adapts to the provided settings (budget, number of workers, parametrization) and should therefore be a good
# default. @TwoPointsDE is excellent in many cases, including very high num_workers. @PortfolioDiscreteOnePlusOne is
# excellent in discrete settings of mixed settings when high precision on parameters is not relevant; it’s possibly a
# good choice for hyperparameter choice. @OnePlusOne is a simple robust method for continuous parameters with
# num_workers < 8. @CMA is excellent for control (e.g. neurocontrol) when the environment is not very noisy (
# num_workers ~50 ok) and when the budget is large (e.g. 1000 x the dimension). @TBPSA is excellent for problems
# corrupted by noise, in particular overparameterized (neural) ones; very high num_workers ok). @PSO is excellent in
# terms of robustness, high num_workers ok. @ScrHammersleySearchPlusMiddlePoint is excellent for super parallel cases
# (fully one-shot, i.e. num_workers = budget included) or for very multimodal cases (such as some of our MLDA
# problems); don’t use softmax with this optimizer. @RandomSearch is the classical random search baseline; don’t use
# softmax with this optimizer.
def choose_optimizer_algoritm(optimizer_algoritm, budget, num_workers, instrum):
    if optimizer_algoritm == 'TwoPointsDE':
        optimizer = ng.optimizers.TwoPointsDE(parametrization=instrum, budget=budget, num_workers=num_workers)
    if optimizer_algoritm == 'NGOpt':
        optimizer = ng.optimizers.NGOpt(parametrization=instrum, budget=budget, num_workers=num_workers)
    if optimizer_algoritm == 'PortfolioDiscreteOnePlusOne':
        optimizer = ng.optimizers.PortfolioDiscreteOnePlusOne(parametrization=instrum, budget=budget,
                                                              num_workers=num_workers)
    if optimizer_algoritm == 'OnePlusOne':
        optimizer = ng.optimizers.OnePlusOne(parametrization=instrum, budget=budget, num_workers=num_workers)
    if optimizer_algoritm == 'CMA':
        optimizer = ng.optimizers.CMA(parametrization=instrum, budget=budget, num_workers=num_workers)
    if optimizer_algoritm == 'TBPSA':
        optimizer = ng.optimizers.TBPSA(parametrization=instrum, budget=budget, num_workers=num_workers)
    if optimizer_algoritm == 'PSO':
        optimizer = ng.optimizers.PSO(parametrization=instrum, budget=budget, num_workers=num_workers)
    if optimizer_algoritm == 'ScrHammersleySearchPlusMiddlePoint':
        optimizer = ng.optimizers.ScrHammersleySearchPlusMiddlePoint(parametrization=instrum, budget=budget,
                                                                     num_workers=num_workers)
    if optimizer_algoritm == 'RandomSearch':
        optimizer = ng.optimizers.RandomSearch(parametrization=instrum, budget=budget, num_workers=num_workers)

    return optimizer


def optimize_metric(metric_return, metrics_values):
    result = []
    for key, value in metrics_values.items():
        for m in metric_return:
            if m in key:
                if 'rsq' in metric_return:
                    result.append(-value)
                else:
                    result.append(value)
    return result



from cassandra.data.dataAnalysis.correlation import *
from cassandra.data.dataProcessing.prophet import *
from cassandra.data.featureSelection.ols import *
from cassandra.data.dataAnalysis.exploration.utils import show_media, show_no_media, create_df, define_features
from cassandra.data.featureSelection.multicollinearity import check_multicollinearity
import pandas as pd
import time
from IPython.core.display import display
from cassandra.model.modelEvaluation.plot import show_seasonality


def exploration(csv, name_date_column, name_target_column, dict_columns_aggregate={}, window_start='',
                window_end='', national_holidays_abbreviation='IT', future_dataframe_periods=14, plot_prophet=False,
                weekly_seasonality=True, yearly_seasonality=True, daily_seasonality=True, seasonality_mode='additive',
                media_split=True, show_graph_media=True, show_graph_no_media=True, method_corr='pearson',
                min_periods_corr=1, plot_for_line=5, show_only_one='',
                constant_ols=True):
    # create csv
    if dict_columns_aggregate:
        df = create_df(csv, dict_columns_aggregate)
    else:
        df = csv

    display(Markdown(f"# Dataset"))
    display(df)

    # create df with seasonality variables
    new_df = prophet(df=df, name_date_column=name_date_column, name_target_column=name_target_column,
                     window_start=window_start, window_end=window_end,
                     national_holidays_abbreviation=national_holidays_abbreviation,
                     future_dataframe_periods=future_dataframe_periods, plot_prophet=plot_prophet,
                     weekly_seasonality=weekly_seasonality, yearly_seasonality=yearly_seasonality,
                     daily_seasonality=daily_seasonality, seasonality_mode=seasonality_mode)

    # define variables and show graph each other
    if media_split:
        everything, media, organic, price, seasonality, no_media, no_organic, no_price, no_seasonality, media_google, media_fb, media_others = define_features(
            df=new_df, media_split=media_split)

        if show_graph_media:
            show_media(new_df, name_date_column, name_target_column, media, media_google, media_fb, media_others)
        if show_graph_no_media:
            show_no_media(new_df, name_date_column, name_target_column, organic, price, seasonality)

    else:
        everything, media, organic, price, seasonality, no_media, no_organic, no_price, no_seasonality = define_features(
            df=new_df, media_split=media_split)
        if show_graph_media:
            show_media(new_df, name_date_column, name_target_column, media)
        if show_graph_no_media:
            show_no_media(new_df, name_date_column, name_target_column, organic, price, seasonality)

    pd.set_option("display.max_rows", None)
    pd.set_option("display.max_columns", None)
    new_df.fillna(value=0, inplace=True)
    # df.fillna(method = 'bfill', inplace=True)
    # df.fillna(method = 'ffill', inplace=True)

    display(Markdown(f"# Modeling Seasonality"))
    seasonality_df = show_seasonality(new_df, seasonality, name_date_column)

    display(Markdown(f"## Check Pearson Correlation Metric to check various possible correlations"))
    display(new_df.corr(method=method_corr, min_periods=min_periods_corr))

    display(Markdown(
        f"## Show through a scatterplot the relationship between each input variable and {name_target_column}"))
    show_relationship_between_variables(new_df, plot_for_line=plot_for_line, show_only_one=show_only_one)

    # Features Selection
    X = new_df[everything]
    y = new_df[name_target_column]

    # OLS model
    display(Markdown(f"## OLS Model"))
    result_ols, model_ols = ols(X, y, constant=constant_ols)

    # check for multicollinearity
    display(Markdown(f"## Check for multicollinearity"))
    check_multicollinearity(X)

    pd.set_option("display.max_rows", 20)
    pd.set_option("display.max_columns", 20)

    return new_df



from cassandra.data.dataAnalysis.correlation import *
from cassandra.data.dataAnalysis.plot import *
import pandas as pd


# Example how create a dict_columns_aggregate dict_columns_aggregate = { 'keywords_impressions' : [
# 'adopta_albero_impressions', 'regalar_arbol_impressions', 'regala_arbol_impressions', 'plantar_arbol_impressions',
# 'treedom_impressions'] }
def create_df(df, dict_columns_aggregate={}):
    for key, value in dict_columns_aggregate.items():
        # Aggregate columns into one for each key
        df[key] = df[value].sum(axis=1)

        # Drop un-used columns
        df.drop(value, axis=1, inplace=True)
        df.fillna(value=0)

    return df


def define_features(df, media_split=True):
    media = [col for col in df.columns if '_spend' in col]

    organic = [col for col in df.columns if '_org' in col]

    seasonality = []
    if 'trend' in df.columns:
        seasonality.append('trend')
    if 'holidays' in df.columns:
        seasonality.append('holidays')
    if 'yearly' in df.columns:
        seasonality.append('yearly')
    if 'monthly' in df.columns:
        seasonality.append('monthly')
    if 'weekly' in df.columns:
        seasonality.append('weekly')
    if 'daily' in df.columns:
        seasonality.append('daily')

    price = [col for col in df.columns if 'avg' in col]

    media_google = [col for col in media if col.startswith('g_')]

    media_fb = [col for col in media if col.startswith('fb_')]

    #media_am = [col for col in media if col.startswith('am_')]

    media_others = [col for col in media if not col.startswith('g_') if not col.startswith('fb_')]

    everything = media + organic + price + seasonality
    no_price = media + organic + seasonality
    no_seasonality = media + organic + seasonality
    no_organic = media + price + seasonality
    no_media = organic + price + seasonality

    if media_split:
        return everything, media, organic, price, seasonality, no_media, no_organic, no_price, no_seasonality, media_google, media_fb, media_others
    else:
        return everything, media, organic, price, seasonality, no_media, no_organic, no_price, no_seasonality


def show_media(df, name_date_column, name_target_column, media, media_google=[], media_fb=[], media_others=[]):
    if media_google:
        #Google media graph
        display(Markdown(f"# Google media"))
        display(Markdown(f"### Google media over time"))
        fig_media_google = px.line(df, x=name_date_column, y=media_google)
        fig_media_google.show()
        display(Markdown(f"### Google media trendline"))
        show_trendline_curves(media_google, df, name_target_column)

    if media_fb:
        #Facebook media graph
        display(Markdown(f"# Facebook media"))
        display(Markdown(f"### Facebook media over time"))
        fig_media_fb = px.line(df, x=name_date_column, y=media_fb)
        fig_media_fb.show()
        display(Markdown(f"### Facebook media trendline"))
        show_trendline_curves(media_fb, df, name_target_column)

    if media_others:
        #Others media graph
        display(Markdown(f"# Others media"))
        display(Markdown(f"### Others media over time"))
        fig_media_others = px.line(df, x=name_date_column, y=media_others)
        fig_media_others.show()
        display(Markdown(f"### Others media trendline"))
        show_trendline_curves(media_others, df, name_target_column)


    #Target vs media
    display(Markdown(f"# Time serie: media and {name_target_column}"))
    show_target_vs_media_spent_graph(df, name_date_column, name_target_column, media)


def show_no_media(df, name_date_column, name_target_column, organic=[], price=[], seasonality=[]):
    if organic:
        # Organic graph
        display(Markdown(f"# Organic"))
        display(Markdown(f"### Organic over time"))
        fig_organic = px.line(df, x=name_date_column, y=organic)
        fig_organic.show()
        display(Markdown(f"### Organic trendline"))
        show_trendline_curves(organic, df, name_target_column)
        # Target vs organic
        display(Markdown(f"# Time serie: organic and {name_target_column}"))
        show_target_vs_media_spent_graph(df, name_date_column, name_target_column, organic)

    if price:
        # Organic graph
        display(Markdown(f"# Price"))
        display(Markdown(f"### Price over time"))
        fig_price = px.line(df, x=name_date_column, y=price)
        fig_price.show()
        display(Markdown(f"### Price trendline"))
        show_trendline_curves(price, df, name_target_column)
        # Target vs price
        display(Markdown(f"# Time serie: price and {name_target_column}"))
        show_target_vs_media_spent_graph(df, name_date_column, name_target_column, price)

    if seasonality:
        # Seasonality graph
        display(Markdown(f"# Seasonality"))
        display(Markdown(f"### Seasonality over time"))
        fig_seasonality = px.line(df, x=name_date_column, y=seasonality)
        fig_seasonality.show()
        display(Markdown(f"### Seasonality trendline"))
        show_trendline_curves(seasonality, df, name_target_column)
        # Target vs seasonality
        display(Markdown(f"# Time serie: seasonality and {name_target_column}"))
        show_target_vs_media_spent_graph(df, name_date_column, name_target_column, seasonality)


